.. This (-*- rst -*-) format file contains commonly used link targets
   and name substitutions.  It may be included in many files,
   therefore it should only contain link targets and name
   substitutions.  Try grepping for "^\.\. _" to find plausible
   candidates for this list.

.. NOTE: reST targets are
   __not_case_sensitive__, so only one target definition is needed for
   nipy, NIPY, Nipy, etc...


.. _DataLad: https://www.datalad.org
.. _DataLad Handbook: http://handbook.datalad.org/en/latest/
.. _GitHub: https://www.github.com/
.. _git-annex: git-annex.branchable.com/
.. _git: git-scm.com/
.. _OSF: https://www.osf.io/
.. _pip: https://pypi.org/project/pip/
.. _Python: https://www.python.org/
.. _Special Remote: https://git-annex.branchable.com/special_remotes/
.. _Special Remotes: https://git-annex.branchable.com/special_remotes/
